#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # 2A.ml - Timeseries et machine learning
# 
# Série temporelle et prédiction. Module [statsmodels](http://www.statsmodels.org/stable/index.html).

# In[1]:


get_ipython().run_line_magic('matplotlib', 'inline')
import matplotlib.pyplot as plt


# Les séries temporelles diffèrent des problèmes de machine learning classique car les observations ne sont pas indépendantes. Ce notebook présente quelques traitements classiques et une façon assez simple d'appliquer un traitement de machine learning.

# In[2]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# Je ne vais pas refaire ici un cours sur les séries temporelles. Quelques références pour approfondir :
#    
# * [Différencier (indéfiniment) les séries temporelles ?](https://freakonometrics.hypotheses.org/1876)
# * [Cours de Séries Temporelles - Arthur Charpentier](http://cours.mido.dauphine.fr/st/cours.html)
# * [Modèles de prévision Séries temporelles](http://freakonometrics.free.fr/M1_TS_2015.pdf)
# * [COURS DE SERIES TEMPORELLES THEORIE ET APPLICATIONS](https://perso.univ-rennes1.fr/arthur.charpentier/TS1.pdf)

# ## Données

# Prenons comme exemple une série venant de [Google Trend](https://www.google.com/trends/), par exemple le terme [macron](https://www.google.com/trends/explore?geo=FR&q=macron) pris sur la France.

# In[3]:


from ensae_teaching_cs.data import google_trends
t = google_trends("macron") # export CSV, donc pas à jour
import pandas
df = pandas.read_csv(t, sep=",")
df.tail()


# In[4]:


df.plot(x="Week", y="macron", figsize=(14,4))


# ## Période, tendance

# On enlève tout ce qui précède 08/2014 car la série se comporte de façon différente. Monsieur Macron n'était probablement pas ministre à avant cette date.

# In[5]:


df[df.Week >= "2014-08"].plot(x="Week", y="macron", figsize=(14,4))


# On peut étudier les corrélations entre les différentes dates. Le module le plus approprié est [statsmodels](http://statsmodels.sourceforge.net/devel/index.html).

# In[6]:


data = df[df.Week >= "2014-08"].copy()


# La série ne montre pas de période évidente. Regardons la tendance.

# In[7]:


from statsmodels.tsa.tsatools import detrend
notrend = detrend(data.macron)
data["notrend"] = notrend
data.plot(x="Week", y=["macron", "notrend"], figsize=(14,4))


# On vérifie pour la période.

# In[8]:


from statsmodels.tsa.stattools import acf
cor = acf(data.notrend)
cor


# In[9]:


plt.plot(cor)


# ## Auto-corrélation

# On recommence sur la série [chocolat](https://www.google.com/trends/explore?geo=FR&q=chocolat).

# In[10]:


choco = pandas.read_csv(google_trends("chocolat"), sep=",")
choco.plot(x="Week", y="chocolat", figsize=(14,4))


# In[11]:


choco["notrend"] = detrend(choco.chocolat)
cor = acf(choco.notrend)
plt.plot(cor)


# On regarde plus souvent les [auto-corrélations partielles](https://en.wikipedia.org/wiki/Partial_autocorrelation_function).

# In[12]:


from statsmodels.tsa.stattools import pacf
pcor = pacf(choco.notrend)
plt.plot(pcor)


# ## Tendance

# La fonction [detrend](http://statsmodels.sourceforge.net/stable/generated/statsmodels.tsa.tsatools.detrend.html#statsmodels.tsa.tsatools.detrend) retourne la tendance. On l'obtient en réalisant une régression linéaire de $Y$ sur le temps $t$.

# In[13]:


from statsmodels.api import OLS
import numpy
y = data.macron
X = numpy.ones((len(y), 2))
X[:,1] = numpy.arange(0,len(y))
reg = OLS(y,X)
results = reg.fit()
results.params


# In[14]:


from statsmodels.graphics.regressionplots import abline_plot
fig = abline_plot(model_results=results)
ax = fig.axes[0]
ax.plot(X[:,1], y, 'b')
ax.margins(.1)


# ## Prédictions linéaires

# On sort ici du cadre linéaire pour utiliser le machine learning non linéaire pour faire de la prédiction. Pour éviter les trop gros problèmes de tendance, on travaillera sur la série différenciée. En théorie, il faudrait différencier jusqu'à ce qu'on enlève la tendance (voir [ARIMA](https://en.wikipedia.org/wiki/Autoregressive_integrated_moving_average)).

# In[15]:


from ensae_teaching_cs.data import google_trends
df = pandas.read_csv(google_trends("chocolat"), sep=",")
df.plot(x="Week", y="chocolat", figsize=(14,4))


# On calcule la série différenciée. Quelques astuces. [pandas](http://pandas.pydata.org/) utilise les indices par défaut pour faire des opérations sur les colonnes. C'est pourquoi il faut convertir les tableaux extraits du dataframe en [numpy array](http://docs.scipy.org/doc/numpy/reference/generated/numpy.array.html). **N'oubliez pas ce détail. Cela peut devenir très agaçant**. On peut aussi utiliser la fonction [diff](http://pandas.pydata.org/pandas-docs/version/0.17.0/generated/pandas.DataFrame.diff.html).

# In[16]:


df["diff"] = numpy.nan
df.loc[1:, "diff"] = (df.iloc[1:, 1].values - df.iloc[:len(df)-1, 1].values)
pandas.concat([df.head(n=3), df.tail(n=3)])


# In[17]:


df.plot(x="Week", y="diff", figsize=(14,4))


# Sans information, la meilleure prédiction est la valeur de la veille (où celle dans le passé à l'horizon de prédiction considéré). Dans notre cas, c'est simplement la variance de la série différenciée :

# In[18]:


(df["diff"].apply(lambda x:x**2).sum()/len(df))**0.5


# In[19]:


from statsmodels.tsa.arima_model import ARMA
arma_mod = ARMA(df["diff"].dropna().values, order=(8, 1))
res = arma_mod.fit()
res.params


# **à finir** train/test / prédiciton

# ## Machine learning

# On souhaite utiliser une random forest pour faire de la prédiction. On créé la matrice avec les séries décalées.

# In[20]:


from statsmodels.tsa.tsatools import lagmat
lag = 8
X = lagmat(df["diff"], lag)
lagged = df.copy()
for c in range(1,lag+1):
    lagged["lag%d" % c] = X[:, c-1]
pandas.concat([lagged.head(), lagged.tail()])


# On découpe en train/test de façon non aléatoire car c'est une série temporelle.

# In[21]:


xc = ["lag%d" % i for i in range(1,lag+1)]
split = 0.66
isplit = int(len(lagged) * split)
xt = lagged[10:][xc]
yt = lagged[10:]["diff"]
X_train, y_train, X_test, y_test = xt[:isplit], yt[:isplit], xt[isplit:], yt[isplit:]


# On peut maintenant faire du machine learning sur la série décalé.

# In[22]:


from sklearn.linear_model import LinearRegression
clr = LinearRegression()
clr.fit(X_train, y_train)


# In[23]:


from sklearn.metrics import r2_score
r2 = r2_score(y_test.values, clr.predict(X_test))
r2


# In[24]:


plt.scatter(y_test.values, clr.predict(X_test))


# Non linéaire

# In[25]:


from sklearn.ensemble import RandomForestRegressor
clrf = RandomForestRegressor()
clrf.fit(X_train, y_train)


# In[26]:


from sklearn.metrics import r2_score
r2 = r2_score(y_test.values, clrf.predict(X_test))
r2


# C'est mieux.

# In[27]:


plt.scatter(y_test.values, clrf.predict(X_test))


# Le modèle prédit bien les extrêmes, il faudrait sans doute les enlever ou utiliser une échelle logarithmique pour réduire leur importance.

# In[28]: